#pragma once
#include "IRprinter.h"
#include "Instruction.h"
class MovInst : public BaseInst<MovInst> {
    friend BaseInst<MovInst>;

  protected:
    MovInst(std::shared_ptr<Value> to, std::shared_ptr<Value> from, BasicBlock *bb)
        : BaseInst<MovInst>(Type::get_void_type(bb->get_module()), Instruction::mov, 2, bb) {
        set_operand(0, to);
        set_operand(1, from);
    }

  public:
    // physical register?
    static std::shared_ptr<MovInst> create_mov(std::shared_ptr<Value> to, std::shared_ptr<Value> from, BasicBlock *bb) {
        return create(to, from, bb);
    }
    auto get_from() { return get_operand(1); }
    auto get_to() { return get_operand(0); }
    virtual std::string print() override;
};

std::string MovInst::print() {
    std::string lir;
    lir += get_module()->get_instr_op_name(get_instr_type());
    lir += " ";
    lir += this->get_operand(0)->get_type()->print();
    lir += " ";
    lir += print_as_op(get_operand(0).get(), false);
    lir += " ";
    lir += this->get_operand(1)->get_type()->print();
    lir += " ";
    lir += print_as_op(get_operand(1).get(), false);
    return lir;
}
