#pragma once
#include "Function.h"
#include "Pass.hh"
#include "logging.hpp"

#include <deque>
#include <unordered_map>
#include <unordered_set>

/**
 * @brief 全局公共子表达式消除，应该在局部公共子表示消除后调用
 * 原理参见 高级编译器设计与实现-全局公共子表达式删除
 */
class GlobalCommonExpression : public Pass {
  public:
    GlobalCommonExpression(Module *m) : Pass(m) {}

    void run() {
        auto func_list = m_->get_functions();
        for (auto func : func_list) {
            if (func->get_num_basic_blocks() == 0)
                continue;
            process_func(func);
        }

        LOG_INFO << "local common expression process " << ins_count << " instructions";
    }

  private:
    void process_func(std::shared_ptr<Function> func) {
        // 基本流程
        // 1. 处理出函数的数据流
        // 2. 求解数据流方程
        // 3. 进行全局公共子表达式删除
    }

    int ins_count{0};
};
