#ifndef PASS_HH
#define PASS_HH

#include "Module.h"
#include "logging.hpp"

#include <vector>

using std::shared_ptr;
using std::vector;

class Pass {
  public:
    Pass(Module *m) : m_(m) {}
    virtual void run() = 0;
    virtual ~Pass() = default;

  protected:
    Module *m_;
};

class IRCheck : public Pass {
  public:
    IRCheck(Module *m) : Pass(m) {}
    ~IRCheck(){};
    void run() override;

  private:
    void check_parent();
    void check_phi_position();
    void check_ret_br_position();
    void check_terminate();
    void check_pred_succ();
    void check_entry();
    void check_use_list();
    void check_operand_exit();
};

class PassManager {
  private:
    Module *m_;
    IRCheck ir_check;

  public:
    vector<std::pair<std::shared_ptr<Pass>, bool>> pass_list;
    PassManager(Module *m) : m_(m), ir_check(m) {}

    template <typename T>
    void add_pass(bool print_IR = false) {
        pass_list.push_back({std::make_shared<T>(m_), print_IR});
    }

    void run() {
        // try {
        //     ir_check.run();
        // } catch (...) {
        //     LOG_ERROR << "IRCheck ERROR after SYSYBuilder";
        //     // exit(-1);
        // }
        for (auto [pass, print_IR] : pass_list) {
            pass->run();
            if (print_IR)
                std::cout << m_->print() << std::endl;
            try {
                ir_check.run();
            } catch (...) {
                exit(203);
                LOG_ERROR << "IRCheck ERROR after pass " << typeid(*pass).name();
                // exit(-1);
            }
        }
    }
};
#endif
