#include "Constant.h"

#include "Module.h"

#include <iostream>
#include <sstream>

std::shared_ptr<ConstantInt> ConstantInt::get(int val, Module *m) {
    return std::shared_ptr<ConstantInt>(new ConstantInt(Type::get_int32_type(m), val));
}
std::shared_ptr<ConstantInt> ConstantInt::get(bool val, Module *m) {
    return std::shared_ptr<ConstantInt>(new ConstantInt(Type::get_int1_type(m), val ? 1 : 0));
}
std::shared_ptr<ConstantInt> ConstantInt::get_i64(long val, Module *m) {
    return std::shared_ptr<ConstantInt>(new ConstantInt(Type::get_int64_type(m), val));
}

std::string ConstantInt::print() {
    std::string const_ir;
    Type *ty = this->get_type();
    if (ty->is_integer_type() && static_cast<IntegerType *>(ty)->get_num_bits() == 1) {
        // int1
        const_ir += (this->get_value() == 0) ? "false" : "true";
    } else {
        // int32
        const_ir += std::to_string(this->get_value());
    }
    return const_ir;
}

ConstantArray::ConstantArray(ArrayType *ty, const std::vector<std::shared_ptr<Constant>> &val)
    : Constant(ty, "", val.size()) {
    for (int i = 0; i < val.size(); i++)
        set_operand(i, std::static_pointer_cast<Value>(val[i]));
    this->const_array.assign(val.begin(), val.end());
}

std::shared_ptr<Constant> ConstantArray::get_element_value(int index) { return this->const_array[index]; }

std::shared_ptr<ConstantArray> ConstantArray::get(ArrayType *ty, const std::vector<std::shared_ptr<Constant>> &val) {
    return std::shared_ptr<ConstantArray>(new ConstantArray(ty, val));
}

std::string ConstantArray::print() {
    std::string const_ir;
    const_ir += this->get_type()->print();
    const_ir += " ";
    const_ir += "[";
    for (int i = 0; i < this->get_size_of_array(); i++) {
        std::shared_ptr<Constant> element = get_element_value(i);
        if (!std::dynamic_pointer_cast<ConstantArray>(get_element_value(i))) {
            const_ir += element->get_type()->print();
        }
        const_ir += " ";
        const_ir += element->print();
        if (i < this->get_size_of_array() - 1) {
            const_ir += ", ";
        }
    }
    const_ir += "]";
    return const_ir;
}

std::string ConstantArray::print_initializer() {
    std::string const_ir;
    // const_ir += this->get_type()->print();
    // const_ir += " ";
    const_ir += "[";
    for (int i = 0; i < this->get_size_of_array(); i++) {
        std::shared_ptr<Constant> element = get_element_value(i);
        if (!std::dynamic_pointer_cast<ConstantArray>(get_element_value(i))) {
            const_ir += element->get_type()->print();
        }
        const_ir += " ";
        const_ir += element->print();
        if (i < this->get_size_of_array() - 1) {
            const_ir += ", ";
        }
    }
    const_ir += "]";
    return const_ir;
}

std::shared_ptr<ConstantFP> ConstantFP::get(float val, Module *m) {
    return std::shared_ptr<ConstantFP>(new ConstantFP(Type::get_float_type(m), val));
}

std::string ConstantFP::print() {
    std::stringstream fp_ir_ss;
    std::string fp_ir;
    double val = this->get_value();
    fp_ir_ss << "0x" << std::hex << *(uint64_t *)&val << std::endl;
    // std::cout << "0d" << std::dec << *(uint64_t *)&val << std::endl;
    fp_ir_ss >> fp_ir;
    return fp_ir;
}

std::shared_ptr<ConstantZero> ConstantZero::get(Type *ty, Module *m) {
    return std::shared_ptr<ConstantZero>(new ConstantZero(ty));
}

std::string ConstantZero::print() { return "zeroinitializer"; }
