#include "Value.h"

#include "Constant.h"
#include "Instruction.h"
#include "Type.h"
#include "User.h"

#include <cassert>

Value::Value(Type *ty, const std::string &name) : type_(ty), name_(name) {}

void Value::add_use(Value *val, unsigned arg_no) { use_list_.push_back(Use(val, arg_no)); }

std::string Value::get_name() const { return name_; }

void Value::replace_all_use_with(std::shared_ptr<Value> new_val) {
    for (auto use : use_list_) {
        auto val = dynamic_cast<User *>(use.val_);
        assert(val && "new_val is not a user");
        auto rand = val->get_operand(use.arg_no_);
        if (rand.is_weak() and not std::dynamic_pointer_cast<Constant>(new_val))
            val->set_operand(use.arg_no_, std::weak_ptr(new_val));
        else
            val->set_operand(use.arg_no_, new_val);
    }
    use_list_.clear();
}

void Value::remove_use(Value *val) {
    auto is_val = [val](const Use &use) { return use.val_ == val; };
    use_list_.remove_if(is_val);
}
