#include "IVReduction.hh"

#include "utils.hh"

void IVReduction::run() {
    // debug
    // std::cout << m_->print();
    loop_info.run();
    for (auto loop : loop_info.get_loops()) {
        loop.find_dependent_IV();
        for (auto iv : loop.get_ind_vars()) {
            if (iv->type == ivtype::dependent) {
                auto const_c1 = utils::get_const_int_val(iv->c1.get());
                auto const_c2 = utils::get_const_int_val(iv->c2.get());
                if (const_c1.has_value() && const_c1.value() == 1)
                    continue;
                /// for (int basic = init_b; basic < final_b; basic += step_b) {
                ///     dependent = basic * c1 + c2;
                /// }
                /// initial value of dependent: init_b * c1 + c2 - step
                /// step size of dependent: step_b * c1
                /// final value of dependent: final_b * c1 + c2
                auto step = utils::mul_i32(iv->c1, iv->basis->step_val, m_);
                auto init_mul_part = utils::mul_i32(iv->c1, iv->basis->initial_val, m_);
                auto init_mul_sub_step = utils::sub_i32(init_mul_part, step, m_);
                std::shared_ptr<Value> init;
                if (const_c2.has_value() && const_c2.value() == 0)
                    init = init_mul_sub_step;
                else
                    init = utils::add_i32(init_mul_sub_step, iv->c2, m_);
                // auto final_mul_part = utils::mul_i32(iv->c1, iv->basis->final_val, m_);
                // std::shared_ptr<Value> finalv;
                // if (const_c2.has_value() && const_c2.value() == 0)
                //     finalv = final_mul_part;
                // else
                //     finalv = utils::add_i32(final_mul_part, iv->c2, m_);
                // auto iv_bb = iv->inst->get_parent();
                loop.insert_into_preheader(step);
                loop.insert_into_preheader(init_mul_part);
                loop.insert_into_preheader(init_mul_sub_step);
                if (init != init_mul_sub_step)
                    loop.insert_into_preheader(init);
                auto new_iv_phi = PhiInst::create_phi(Type::get_int32_type(m_));
                new_iv_phi->add_phi_pair_operand(init, loop.get_preheader());
                auto new_step_inst = std::static_pointer_cast<Instruction>(utils::add_i32(new_iv_phi, step, m_));
                new_iv_phi->add_phi_pair_operand(std::weak_ptr<Value>(new_step_inst),
                                                 std::static_pointer_cast<BasicBlock>(loop.get_end()));
                loop.insert_phi_at_begin(new_iv_phi);

                // 新指令插入原指令之前
                // for (auto inst_itr = iv_bb->get_instructions().begin(); inst_itr != iv_bb->get_instructions().end();
                //      inst_itr++) {
                //     if (inst_itr->get() == iv->inst.get()) {
                //         iv_bb->get_instructions().insert(inst_itr, new_step_inst);
                //         new_step_inst->set_parent(iv_bb);
                //         break;
                //     }
                // }

                // 新指令插入循环开始处
                loop.insert_at_begin(new_step_inst);

                iv->inst->replace_all_use_with(new_step_inst);
                // iv->inst->remove_use_of_ops();
                iv->inst->get_parent()->delete_instr(iv->inst);
                LOG_DEBUG << "weaken dependent iv: " << iv->inst->get_name() << " in func "
                          << loop.get_base()->get_parent()->get_name() << " bb " << iv->inst->get_parent()->get_name();
            }
        }
    }
    // debug
    // std::cout << m_->print();
}