#include "utils.hh"

namespace utils {
std::shared_ptr<Value> to_i32(std::shared_ptr<Value> v, Module *m) {
    auto i32_type = Type::get_int32_type(m);
    auto fp_type = Type::get_float_type(m);
    if (v->get_type() == i32_type)
        return v;
    if (auto v_c = get_const_float_val(v.get()); v_c.has_value())
        return ConstantInt::get(static_cast<int>(v_c.value()), m);
    return SiToFpInst::create_sitofp(v, fp_type, nullptr);
}

std::shared_ptr<Value> add_i32(std::shared_ptr<Value> v1, std::shared_ptr<Value> v2, Module *m) {
    auto i32_type = Type::get_int32_type(m);
    if (!(v1->get_type() == i32_type && v2->get_type() == i32_type)) {
        exit(204);
        LOG_ERROR << "both ops should be i32 type";
        return nullptr;
    }
    auto v1_c = get_const_int_val(v1.get());
    auto v2_c = get_const_int_val(v2.get());
    if (v1_c.has_value() && v2_c.has_value())
        return ConstantInt::get(v1_c.value() + v2_c.value(), m);
    // if (v1_c.has_value() && v1_c.value() == 0)
    //     return v2;
    // if (v2_c.has_value() && v2_c.value() == 0)
    //     return v1;
    return BinaryInst::create(i32_type, Instruction::add, v1, v2, nullptr);
}

std::shared_ptr<Value> sub_i32(std::shared_ptr<Value> v1, std::shared_ptr<Value> v2, Module *m) {
    auto i32_type = Type::get_int32_type(m);
    if (!(v1->get_type() == i32_type && v2->get_type() == i32_type)) {
        exit(205);
        LOG_ERROR << "both ops should be i32 type";
        return nullptr;
    }
    auto v1_c = get_const_int_val(v1.get());
    auto v2_c = get_const_int_val(v2.get());
    if (v1_c.has_value() && v2_c.has_value())
        return ConstantInt::get(v1_c.value() - v2_c.value(), m);
    // if (v2_c.has_value() && v2_c.value() == 0)
    //     return v1;
    return BinaryInst::create(i32_type, Instruction::sub, v1, v2, nullptr);
}

std::shared_ptr<Value> mul_i32(std::shared_ptr<Value> v1, std::shared_ptr<Value> v2, Module *m) {
    auto i32_type = Type::get_int32_type(m);
    if (!(v1->get_type() == i32_type && v2->get_type() == i32_type)) {
        exit(206);
        LOG_ERROR << "both ops should be i32 type";
        return nullptr;
    }
    auto v1_c = get_const_int_val(v1.get());
    auto v2_c = get_const_int_val(v2.get());
    if (v1_c.has_value() && v2_c.has_value())
        return ConstantInt::get(v1_c.value() * v2_c.value(), m);
    // if (v1_c.has_value() && v1_c.value() == 1)
    //     return v2;
    // if (v2_c.has_value() && v2_c.value() == 1)
    //     return v1;
    return BinaryInst::create(i32_type, Instruction::mul, v1, v2, nullptr);
}
}  // namespace utils
